/*
 Run From Clipboard++ (RFC++)
 Description:
   RFC++ is a small utility for running the contents of the clipboard.
   If the first try doesn't succeed then it will try again with all newlines
   removed from the input.  If the command is still isn't successfully executed
   then a form with the input will be shown for user to manually edit and try
   again.

 Change History:
   2007-10-09 *fa - Fixed label redrawing when the windows was resized.
                  - Cleaned up code.
				  - Added special handling for the case when ShellExecuteEx
				    does not set last error correctly.
				  - Added hyperlinking functionality to the bottom static text.
   2007-10-31 *fa - Upgraded version 
*/
#define WIN32_LEAN_AND_MEAN  // Make it as small as possible.

#include <windows.h>
#include <iostream>
#include <ShellAPI.h>
#include "resource.h"

//#include <commctrl.h>

// Identifiers
#define IDE_COMMAND 1
#define IDS_HEADING 2
#define IDB_RUN     3
#define IDS_INFO    4
#define IDS_ERROR   5
#define IDT_REFRESH 6  // Identifier for the timer used for blinking the label.

#define PROP_ORIGINAL_PROC		TEXT("_Hyperlink_Original_Proc_")
#define PROP_STATIC_HYPERLINK	TEXT("_Hyperlink_From_Static_")

// Global variables.
HWND hWnds[5];
HWND mainHwnd;
HFONT hfLinkHover;
HFONT hfLinkNormal;
HCURSOR hCursorLinkHover;

int blinkTimerInterval = 150;  // Interval between each lable blink.
const char * last_error;       // Variable used to store the label text between
                               // each blink.
int counter = 0;               // Counter for the number of blinks.

// Default and minimum height and widths of the main window.
int minAndDefaultWidth  = 310;
int minAndDefaultHeight = 300;


// Get's the last Win32 error.
LPCTSTR GetLastStringError() {
	// http://msdn2.microsoft.com/en-us/library/ms680582.aspx
	LPTSTR lpMsgBuf;
	DWORD dw = GetLastError();
	DWORD success = FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL, dw, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR)&lpMsgBuf, 0, NULL);
	//LocalFree(lpMsgBuf);  // Don't worry about this.

	// If the message formatted successfully then return that, otherwise return
	// the error that FormatMessage had.
	if (success) return lpMsgBuf;
	else return GetLastStringError();
}

/*
 Displays the last error in a message box. This function is only used for
 debugging.
*/
void DisplayLastError() {
	MessageBox(NULL, GetLastStringError(), "Error", MB_OK | MB_ICONERROR);
}

/*
 Removes the specified character from the given source. This function is used
 for stripping newlines.
*/
char * RemoveCharacter(char * source, const char * character) {
	char * removed = new char [strlen(source)];
	strcpy(removed, "");
	char * result = strtok(source, character);
	while (result != NULL) {
		removed = strcat(removed, result);
		result = strtok(NULL, character);
	}
	return removed;
}

// Returns another version with the newlines removed.
LPCTSTR RemoveNewlines(LPCTSTR command) {
	// If there are no new lines then return the original version.
	if (strchr(command, '\n') == NULL && strchr(command, '\r') == NULL) { return command; }

	size_t len = strlen(command);
	char * copiedCommand = new char[len+1];
	strcpy(copiedCommand, command);
	char * removed;
	removed = RemoveCharacter((char *)copiedCommand, "\r");
	removed = RemoveCharacter(removed, "\n");
	return (LPCTSTR)removed;
}

// Executes the specified command.
bool ExecuteCommand(LPCTSTR command) {
	SHELLEXECUTEINFO shInfo = { sizeof(shInfo) };
	shInfo.fMask  = SEE_MASK_FLAG_NO_UI | SEE_MASK_FLAG_DDEWAIT;  // The shell should not show an error.
	shInfo.nShow  = SW_SHOWNORMAL;
	shInfo.lpVerb = NULL;  // "runas" can be used for running in elevated privileges.
	shInfo.lpFile = command;
	shInfo.lpParameters = NULL;
	shInfo.lpDirectory  = NULL;
	shInfo.hwnd     = NULL;
	shInfo.hInstApp = NULL;

	BOOL success = ShellExecuteEx(&shInfo);

	if (success) return true;
	else {
		// Sometimes ShellExecuteEx plays dirty in that it sets hInstApp to an
		// error code less than 32 and GetLastError() returns an unspecified error
		// code. So we have to separately deal with that. Luckily the concerned
		// hInstApp values line up exactly with the system error so we can just
		// cast those values to get a similar GetLastError() behavior.
		if (shInfo.hInstApp < (HINSTANCE)32) {
			SetLastError((DWORD)shInfo.hInstApp);
		    /*
			Below are the possible error codes.
			SE_ERR_FNF = 2              ' file not found
			SE_ERR_PNF = 3              ' path not found
			SE_ERR_ACCESSDENIED = 5     ' access denied
			SE_ERR_OOM = 8              ' out of memory

			The following ones don't line up correctly but for now I don't care
			about these.
			SE_ERR_DLLNOTFOUND = 32
			SE_ERR_SHARE = 26
			SE_ERR_ASSOCINCOMPLETE = 27
			SE_ERR_DDETIMEOUT = 28
			SE_ERR_DDEFAIL = 29
			SE_ERR_DDEBUSY = 30
			SE_ERR_NOASSOC = 31
			*/
		}
		return false;
	}
}

// Runs the specified command with and without newlines.
bool ExecuteCommandWithoutNewlines(LPCTSTR command) {
	if (ExecuteCommand(command)) { return true; }
	LPCTSTR replaced = RemoveNewlines(command);
	if (ExecuteCommand(replaced)) { return true; }
	return false;
}


// Message handler for the link.
LRESULT CALLBACK InfoLnkProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam) {
	switch (Msg) {
		case WM_SETCURSOR:
			SetCursor(hCursorLinkHover);
			return TRUE;
		case WM_LBUTTONDOWN:
			ExecuteCommand("http://www.thebitguru.com/Link?source=RFC%2B%2B");
			return TRUE;
		case WM_MOUSEMOVE:
			if (GetCapture() != hWnds[IDS_INFO])
			{
				SendMessage(hWnds[IDS_INFO], WM_SETFONT, (WPARAM) hfLinkHover, FALSE);
				InvalidateRect(hWnds[IDS_INFO], NULL, FALSE);
				SetCapture(hWnds[IDS_INFO]);
			} else {
				RECT rect;
				GetWindowRect(hWnds[IDS_INFO], &rect);

				POINT pt = { LOWORD(lParam), HIWORD(lParam) };
				ClientToScreen(hWnds[IDS_INFO], &pt);

				if (!PtInRect(&rect, pt))
				{
					SendMessage(hWnds[IDS_INFO], WM_SETFONT, (WPARAM) hfLinkNormal, FALSE);
					InvalidateRect(hWnds[IDS_INFO], NULL, FALSE);
					ReleaseCapture();
				}
			}
			return FALSE;
	}
	return CallWindowProc((WNDPROC)GetProp(hWnd, PROP_ORIGINAL_PROC), hWnd, Msg, wParam, lParam);
}

// Message handler for link parent.
LRESULT CALLBACK InfoLnkParentProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam) {
	switch (Msg) {
		case WM_CTLCOLORSTATIC:
			{
				HDC hdc = (HDC) wParam;
				HWND hwndCtl = (HWND) lParam;

				BOOL fHyperlink = (NULL != GetProp(hwndCtl, PROP_STATIC_HYPERLINK));
				if (fHyperlink)
				{
					LRESULT lr = CallWindowProc((WNDPROC)GetProp(hWnd, PROP_ORIGINAL_PROC),
						hWnd, Msg, wParam, lParam);
					SetTextColor(hdc, RGB(0, 0, 192));
					return lr;
				}
				break;
			}
	}
	return CallWindowProc((WNDPROC)GetProp(hWnd, PROP_ORIGINAL_PROC), hWnd, Msg, wParam, lParam);
}


/*
 This is the function creates all the controls. This function is called by the
 WndProc when we get a WM_CREATE message. 
*/
void CreateControls(HWND parentHwnd) {
	// *** Create the main text box.
	HFONT hfDefault = (HFONT)GetStockObject(DEFAULT_GUI_FONT);
	HWND hEdit;
	HMODULE moduleHandle = GetModuleHandle(NULL);

	// *** Create the hover cursor.
	hCursorLinkHover = LoadCursor(NULL, MAKEINTRESOURCE(IDC_HAND));
	if (hCursorLinkHover == NULL) { hCursorLinkHover = LoadCursor(NULL, MAKEINTRESOURCE(IDC_ARROW)); }

	// *** Create the hyperlink fonts.
	hfLinkNormal = (HFONT)GetStockObject(DEFAULT_GUI_FONT);
	hfLinkHover  = (HFONT)GetStockObject(DEFAULT_GUI_FONT);
	LOGFONT lf;
	GetObject(hfLinkHover, sizeof(lf), &lf);
	lf.lfUnderline = TRUE;
	hfLinkHover = CreateFontIndirect(&lf);

	// *** Create the error font.
	HFONT hfError = CreateFont(16, 0, 0, 0, FW_BOLD, 0, 0, 0, 0, 0, 0, 0, 0, "Arial");

	// *** Create the command text box.
	hEdit = CreateWindowEx(WS_EX_CLIENTEDGE, "EDIT", "Command will be displayed here.", 
		WS_CHILD | WS_VISIBLE | WS_HSCROLL | WS_VSCROLL | ES_MULTILINE | ES_AUTOVSCROLL | ES_AUTOHSCROLL,
		0, 0, 100, 100, parentHwnd, (HMENU)IDE_COMMAND, moduleHandle, NULL);
	if (hEdit == NULL) {
		MessageBox(NULL, "Could not create the edit box.", "Error", MB_OK | MB_ICONERROR);
		return;
	}
	SendMessage(hEdit, WM_SETFONT, (WPARAM)GetStockObject(DEFAULT_GUI_FONT), MAKELPARAM(FALSE, 0));
	hWnds[IDE_COMMAND] = hEdit;
	SetDlgItemText(parentHwnd, 0, "Testing");
	SendMessage(hEdit, WM_SETTEXT, 0, (LPARAM)"Next line");

	// Create the top label.
	HWND hLabel = CreateWindow("STATIC", "Run From Clipboard++", WS_CHILD | WS_VISIBLE | SS_LEFT,
		0, 0, 170, 20, parentHwnd, (HMENU)IDS_HEADING, moduleHandle, NULL);
	HFONT hfHeading = CreateFont(18, 0, 0, 0, FW_BOLD, FALSE, 0, 0, 0, 0, 0, 0, 0, "Arial");
	SendDlgItemMessage(parentHwnd, IDS_HEADING, WM_SETFONT, (WPARAM)hfHeading, TRUE);

	// Create the run button.
	HWND hButton = CreateWindow("BUTTON", "&Run and Exit", WS_CHILD | WS_VISIBLE,
		0, 0, 75, 25, parentHwnd, (HMENU)IDB_RUN, moduleHandle, NULL);
	SendDlgItemMessage(parentHwnd, IDB_RUN, WM_SETFONT, (WPARAM)hfDefault, TRUE);
	hWnds[IDB_RUN] = hButton;

	// Create the error text.
	HWND hError = CreateWindow("STATIC", "This will display the error.",
		WS_CHILD | WS_VISIBLE | SS_LEFT | SS_SUNKEN,
		0, 0,
		250, 40, parentHwnd, (HMENU)IDS_ERROR, moduleHandle, NULL);
	SendDlgItemMessage(parentHwnd, IDS_ERROR, WM_SETFONT, (WPARAM)hfError, TRUE);
	hWnds[IDS_ERROR] = hError;

	// Create the info static text.
	HWND hInfo = CreateWindow("STATIC",
		"RFC++ by Farhan Ahmad.\nFor more information please visit http://www.thebitguru.com\nVersion 1.0",
		WS_CHILD | WS_VISIBLE | SS_LEFT | SS_NOTIFY, 
		0, 0,
		290, 40, parentHwnd, (HMENU)IDS_INFO, moduleHandle, NULL);
	SendDlgItemMessage(parentHwnd, IDS_INFO, WM_SETFONT, (WPARAM)hfLinkNormal, TRUE);
	SetProp(hInfo, PROP_STATIC_HYPERLINK, (HANDLE)1);  // Set the static control handler.
	SetProp(hInfo, PROP_ORIGINAL_PROC, (HANDLE)GetWindowLong(hInfo, GWL_WNDPROC));  // Set the static control handler.
	SetWindowLongPtr(hInfo, GWLP_WNDPROC, (LONG_PTR)InfoLnkProc);
	SetProp(parentHwnd, PROP_ORIGINAL_PROC, (HANDLE)GetWindowLong(parentHwnd, GWL_WNDPROC));  // Set the static control handler.
	SetWindowLongPtr(parentHwnd, GWLP_WNDPROC, (LONG_PTR)InfoLnkParentProc);
	hWnds[IDS_INFO] = hInfo;

	// Send a WM_SIZE message so that the resize routine can arrange the controls correctly.
	SendMessage(parentHwnd, WM_SIZE, 0, 0);
}

/*
 This is called when the window is resized (WM_SIZE).
*/
void OnWindowResize(HWND hWnd) {
	RECT rcClient;
	GetClientRect(hWnd, &rcClient);  // Get the new size.

	SetWindowPos(GetDlgItem(hWnd, IDS_ERROR), NULL, 5, 50, rcClient.right - 10, 40, SWP_NOZORDER);
	SetWindowPos(GetDlgItem(hWnd, IDE_COMMAND), NULL, 5, 95, rcClient.right - 10, rcClient.bottom - 175, SWP_NOZORDER);
	SetWindowPos(GetDlgItem(hWnd, IDS_HEADING), NULL, (rcClient.right - 170)/2, 10, 0, 0, SWP_NOSIZE);
	SetWindowPos(GetDlgItem(hWnd, IDB_RUN), NULL, rcClient.right - 80, rcClient.bottom - 75, 0, 0, SWP_NOSIZE);
	SetWindowPos(GetDlgItem(hWnd, IDS_INFO), NULL, 5, rcClient.bottom - 45, 0, 0, SWP_NOSIZE);

	InvalidateRect(hWnd, &rcClient, false);  // Redraw everything.
}

// Starts the error blink timer.
void StartErrorBlinkTimer() {
	SetTimer(mainHwnd, IDT_REFRESH, blinkTimerInterval, NULL);
}

// Command called from WndProc when the user clicks the run button.
void RunCommandFromTextbox(HWND hWnd) {
	LPSTR command = new char [1000];
	GetDlgItemText(hWnd, IDE_COMMAND, command, 1000);
	counter = 0;
	if (strcmp(command, "") == 0) {
		StartErrorBlinkTimer();
		SetDlgItemText(hWnd, IDS_ERROR, "Enter a command first.");
	} else if (ExecuteCommandWithoutNewlines(command)) {
		PostQuitMessage(0);
	} else {
		StartErrorBlinkTimer();
		SetDlgItemText(hWnd, IDS_ERROR, GetLastStringError());
	}
}

/*
 This is function called by Windows when an event has occurred.
*/
LRESULT CALLBACK WndProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam) {
	switch(Msg) {
		case WM_CREATE:
			CreateControls(hWnd); // Create the controls.
			StartErrorBlinkTimer();
			break;
		case WM_DESTROY:
			KillTimer(hWnd, IDT_REFRESH);
			PostQuitMessage(0);
			break;
		case WM_TIMER:
			{
				counter++;
				LPSTR error = new char [1000];
				GetDlgItemText(hWnd, IDS_ERROR, error, 1000);
				if (strcmp(error, "") == 0){ 
					SetDlgItemText(hWnd, IDS_ERROR, last_error);
				} else {
					last_error = error;
					SetDlgItemText(hWnd, IDS_ERROR, "");
				}
				if (counter > 10) {
					SetDlgItemText(hWnd, IDS_ERROR, last_error);
					KillTimer(hWnd, IDT_REFRESH);
				}
			}
			break;
		case WM_COMMAND:
			if (LOWORD(wParam) == IDB_RUN) {
				RunCommandFromTextbox(hWnd);
				return TRUE;
			} else if (LOWORD(wParam) == IDA_RUN_COMMAND) {  // ALT+R for running the command.
				RunCommandFromTextbox(hWnd);
				return TRUE;
			} else if (LOWORD(wParam) == IDA_CANCEL) {       // ESCAPE for exiting.
				PostQuitMessage(0);
				return TRUE;
			} 
			//else if (LOWORD(wParam) == IDS_INFO) {         // User clicked on the link.
			//	ExecuteCommand("http://www.thebitguru.com/staticLinkHandler/?source=RFC++");
			//	return TRUE;
			//} 
			else {
				return ::DefWindowProc(hWnd, Msg, wParam, lParam);
			}
			break;
		case WM_SIZE:
			OnWindowResize(hWnd);
			break;
		case WM_GETMINMAXINFO:   // Restrict the minimum window size to 300x300.
			{  // Declare 'p' in a block.
				MINMAXINFO * p = (MINMAXINFO *)lParam;
				p->ptMinTrackSize.x = minAndDefaultWidth;
				p->ptMinTrackSize.y = minAndDefaultHeight;
			}
			return TRUE;
			break;
		default:
			return ::DefWindowProc(hWnd, Msg, wParam, lParam);
	}
	return TRUE;
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
				   int nCmdShow)
{
	// Starting to play around with Common Controls to achieve the new themed look.
    //INITCOMMONCONTROLSEX InitCtrlEx;

    //InitCtrlEx.dwSize = sizeof(INITCOMMONCONTROLSEX);
    //InitCtrlEx.dwICC  = ICC_LINK_CLASS;
    //InitCommonControlsEx(&InitCtrlEx);

	MSG Msg;
	LPCTSTR command = NULL;
	// **** Check the clipboard contents first.
	if (OpenClipboard(NULL)) {
		HANDLE hData = GetClipboardData(CF_TEXT);
		command = (LPCTSTR)GlobalLock(hData);
		if (command == NULL || strcmp(command, "") == 0 || strcmp(RemoveNewlines(command), "") == 0) {  // Make sure we have some text.
			MessageBox(NULL, "Clipboard is empty.", "Run From Clipboard++", MB_OK | MB_ICONINFORMATION);
			return 0;
		}
		GlobalUnlock(hData);
		CloseClipboard();
	}

	// **** Run the command and exit if it succeeded
	if (ExecuteCommandWithoutNewlines(command)) { return 0; }
	LPCTSTR error = GetLastStringError();

	// If that still did not work then show a form to the user.
	const char className[] = "RunFromClipboard";
	WNDCLASS wndClass = {sizeof(WNDCLASS)};
	wndClass.style = 0;
	wndClass.cbClsExtra = 0;
	wndClass.cbWndExtra = 0;
	//wndClass.hbrBackground = (HBRUSH) (COLOR_WINDOW+1);
	wndClass.hbrBackground = (HBRUSH) (COLOR_BTNFACE+1);
	wndClass.lpszClassName = "RunFromClipboard";
	wndClass.hInstance = hInstance;
	wndClass.lpfnWndProc = WndProc;
	wndClass.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_RFCPP));
	wndClass.hCursor = LoadCursor(NULL, IDC_ARROW);

	RegisterClass(&wndClass);

	// Create the main window.
	HWND hWnd = CreateWindow("RunFromClipboard", "Run From Clipboard++", WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, CW_USEDEFAULT,            // Use the default startup position
		minAndDefaultWidth, minAndDefaultHeight, // Startup width and height.
		NULL, NULL, hInstance, NULL);

	if (hWnd == NULL) {
		DisplayLastError();
		MessageBox(NULL, "Unable to create window", "RFC Error", MB_OK | MB_ICONERROR);
		return 1;
	}

	ShowWindow(hWnd, SW_SHOWNORMAL);
	UpdateWindow(hWnd);

	// Set the command and error.
	SetDlgItemText(hWnd, IDE_COMMAND, command);
	SetDlgItemText(hWnd, IDS_ERROR, error);
	mainHwnd = hWnd;

	// *** Bind the accelerator keys
	HACCEL hAccel = ::LoadAccelerators(hInstance, MAKEINTRESOURCE(IDR_ACCELERATORS));

	// Main message loop.
	while(GetMessage(&Msg, NULL, 0, 0) > 0) {
		if (!TranslateAccelerator(hWnd, hAccel, &Msg)) {
			TranslateMessage(&Msg);
			DispatchMessage(&Msg);
		}
	}

	return 0;
}
